/*

  This file is a part of GRALEPS, a point selection utility to help
  create input data for gravitational lens inversion using GRALE.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "backprojectedimage.h"
#include <grale/gravitationallens.h>
#include <grale/imagesdata.h>
#include <grale/gridfunction.h>
#include <grale/constants.h>
#include <serut/dummyserializer.h>
#include <serut/memoryserializer.h>

using namespace grale;

BackProjectedImage::BackProjectedImage()
{
	m_pLens = 0;
}

BackProjectedImage::~BackProjectedImage()
{
	if (m_pLens)
		delete m_pLens;
}


bool BackProjectedImage::init(const ImagesData &imageData, int imageNumber, const GravitationalLens &lens,
                              double Ds, double Dds)
{
	if (m_pLens)
	{
		setErrorString("Already initialized");
		return false;
	}

	int numImages = imageData.getNumberOfImages();

	if (imageNumber < 0 || imageNumber >= numImages)
	{
		setErrorString("Invalid image number");
		return false;
	}
	
	if (!imageData.hasTriangulation())
	{
		setErrorString("Image does not contain a triangulation");
		return false;
	}
	
	std::list<TriangleIndices> triangles;

	if (!imageData.getTriangles(imageNumber, triangles))
	{
		setErrorString(std::string("Unable to obtain triangulation info from image: ") + imageData.getErrorString());
		return false;
	}
	
	int numImagePoints = imageData.getNumberOfImagePoints(imageNumber);

	if (numImagePoints <= 0)
	{
		setErrorString("The number of points in the image must be larger than zero");
		return false;
	}
	
	int numTriangles = triangles.size();

	if (numTriangles <= 0)
	{
		setErrorString("The number of triangles in the image must be larger than zero");
		return false;
	}
	
	m_thetaPoints.resize(numImagePoints);
	m_betaPoints.resize(numImagePoints);
	m_thetaIntens.resize(numImagePoints);
	m_betaIntens.resize(numImagePoints);
	
	for (int i = 0 ; i < numImagePoints ; i++)
	{
		m_thetaPoints[i] = imageData.getImagePointPosition(imageNumber, i);
		if (!imageData.hasIntensities())
			m_thetaIntens[i] = 0;
		else
			m_thetaIntens[i] = imageData.getImagePointIntensity(imageNumber, i);
	}
	
	std::list<TriangleIndices>::const_iterator it;
	int t;

	m_triangles.resize(numTriangles);

	for (t = 0, it = triangles.begin() ; t < numTriangles ; t++, ++it)
		m_triangles[t] = (*it);
	
	// Create a copy of the gravitational lens
	
	serut::DummySerializer dumSer;

	if (!lens.write(dumSer))
	{
		setErrorString("Unable to create a copy of the gravitational lens");
		return false;
	}

	uint8_t *pBuffer = new uint8_t[dumSer.getBytesWritten()];
	serut::MemorySerializer memSer(pBuffer, dumSer.getBytesWritten(), pBuffer, dumSer.getBytesWritten());
	std::string errorString;

	lens.write(memSer);
	
	if (!GravitationalLens::read(memSer, &m_pLens, errorString))
	{
		delete [] pBuffer;
		setErrorString(std::string("Unable to create a copy of the lens: ") + errorString);
		return false;
	}
	
	delete [] pBuffer;

	for (int i = 0 ; i < numImagePoints ; i++)
	{
		Vector2D<double> beta;
		
		if (!m_pLens->traceTheta(Ds, Dds, m_thetaPoints[i], &beta))
		{
			setErrorString(std::string("Unable to project a point back onto the source plane: ") + m_pLens->getErrorString());
			delete m_pLens;
			m_pLens = 0;
			return false;
		}

		m_betaPoints[i] = beta;
		m_betaIntens[i] = m_thetaIntens[i]; // surface brightness is conserved
	}
	
	double minx, miny, maxx, maxy;
	
	minx = m_thetaPoints[0].getX();
	maxx = m_thetaPoints[0].getX();
	miny = m_thetaPoints[0].getY();
	maxy = m_thetaPoints[0].getY();
	
	for (int i = 1 ; i < numImagePoints ; i++)
	{
		minx = MIN(minx, m_thetaPoints[i].getX());
		maxx = MAX(maxx, m_thetaPoints[i].getX());
		miny = MIN(miny, m_thetaPoints[i].getY());
		maxy = MAX(maxy, m_thetaPoints[i].getY());
	}

	m_imageBottomLeft = Vector2D<double>(minx, miny);
	m_imageTopRight = Vector2D<double>(maxx, maxy);
	
	minx = m_betaPoints[0].getX();
	maxx = m_betaPoints[0].getX();
	miny = m_betaPoints[0].getY();
	maxy = m_betaPoints[0].getY();
	
	for (int i = 1 ; i < numImagePoints ; i++)
	{
		minx = MIN(minx, m_betaPoints[i].getX());
		maxx = MAX(maxx, m_betaPoints[i].getX());
		miny = MIN(miny, m_betaPoints[i].getY());
		maxy = MAX(maxy, m_betaPoints[i].getY());
	}

	m_sourceBottomLeft = Vector2D<double>(minx, miny);
	m_sourceTopRight = Vector2D<double>(maxx, maxy);

	for (int i = 0 ; i < m_triangles.size() ; i++)
	{
		int i1 = m_triangles[i].getIndex(0);
		int i2 = m_triangles[i].getIndex(1);
		int i3 = m_triangles[i].getIndex(2);

		Vector2D<double> p1 = m_betaPoints[i1];
		Vector2D<double> p2 = m_betaPoints[i2];
		Vector2D<double> p3 = m_betaPoints[i3];
		double h1 = m_betaIntens[i1];
		double h2 = m_betaIntens[i2];
		double h3 = m_betaIntens[i3];

		m_betaTriangles.push_back(Triangle2DPlus<double>(p1, p2, p3, h1, h2, h3));

		p1 = m_thetaPoints[i1];
		p2 = m_thetaPoints[i2];
		p3 = m_thetaPoints[i3];
		h1 = m_thetaIntens[i1];
		h2 = m_thetaIntens[i2];
		h3 = m_thetaIntens[i3];

		m_thetaTriangles.push_back(Triangle2DPlus<double>(p1, p2, p3, h1, h2, h3));
	}

	m_Ds = Ds;
	m_Dds = Dds;
	
	return true;
}

bool BackProjectedImage::calculateTheta(Vector2D<double> beta, Vector2D<double> &theta, int numIterations, int numIterations2)
{
	if (!m_pLens)
	{
		setErrorString("Not initialized");
		return false;
	}	

	// Find the triangle which contains the 'beta' vector
	
	bool found = false;
	Triangle2D<double> currentThetaTriangle;
	Triangle2D<double> currentBetaTriangle;
	
	for (int i = 0 ; !found && i < m_triangles.size() ; i++)
	{
		Vector2D<double> p1 = m_betaPoints[m_triangles[i].getIndex(0)];
		Vector2D<double> p2 = m_betaPoints[m_triangles[i].getIndex(1)];
		Vector2D<double> p3 = m_betaPoints[m_triangles[i].getIndex(2)];
		
		Triangle2D<double> triangle(p1, p2, p3);

		if (triangle.isInside(beta))
		{
			currentBetaTriangle = triangle;
			
			Vector2D<double> p1 = m_thetaPoints[m_triangles[i].getIndex(0)];
			Vector2D<double> p2 = m_thetaPoints[m_triangles[i].getIndex(1)];
			Vector2D<double> p3 = m_thetaPoints[m_triangles[i].getIndex(2)];
			
			currentThetaTriangle = Triangle2D<double>(p1, p2, p3);
			found = true;
		}
	}

	if (!found)
	{
		setErrorString("The specified source plane point does not lie inside an image triangle");
		return false;
	}

	Vector2D<double> closestPoint = currentThetaTriangle.getCentroid();
	double prevDiff = 1e30;
	bool done = false;
	
	for (int i = 0 ; !done && i < numIterations ; i++)
	{
		Vector2D<double> newTheta = currentThetaTriangle.getCentroid();
		Vector2D<double> newBeta;
	
		if (!m_pLens->traceTheta(m_Ds, m_Dds, newTheta, &newBeta))
		{
			setErrorString(std::string("Unable to trace a point to the source plane: ") + m_pLens->getErrorString());
			return false;
		}

		Vector2D<double> diff = newBeta - beta;

		if (diff.getLength() < prevDiff)
		{
			closestPoint = newTheta;
			prevDiff = diff.getLength();
		}

		const Vector2D<double> *pThetaPoints = currentThetaTriangle.getPoints();
		const Vector2D<double> *pBetaPoints = currentBetaTriangle.getPoints();
		
		bool found = false;
		
		for (int i = 0 ; !found && i < 3 ; i++)
		{
			Vector2D<double> newThetaPoints[3];
			Vector2D<double> newBetaPoints[3];

			for (int j = 0 ; j < 3 ; j++)
			{
				if (j == i)
				{
					newBetaPoints[j] = newBeta;
					newThetaPoints[j] = newTheta;
				}
				else
				{
					newBetaPoints[j] = pBetaPoints[j];
					newThetaPoints[j] = pThetaPoints[j];
				}
			}

			Triangle2D<double> newBetaTriangle(newBetaPoints);

			if (newBetaTriangle.isInside(beta))
			{
				found = true;
				currentBetaTriangle = newBetaTriangle;
				currentThetaTriangle = Triangle2D<double>(newThetaPoints);
			}
		}

		if (!found)
			done = true;
	}

	Vector2D<double> closestBeta;

	if (!m_pLens->traceTheta(m_Ds, m_Dds, closestPoint, &closestBeta))
	{
		setErrorString(std::string("Unable to trace a point to the source plane: ") + m_pLens->getErrorString());
		return false;
	}

	double fraction = m_Dds/m_Ds;
	Vector2D<double> diffBeta = beta - closestBeta;
	Vector2D<double> newTheta = closestPoint;

	prevDiff = diffBeta.getLength();

	for (int i = 0 ; i < numIterations2 ; i++)
	{
		double mxx, myy, mxy;
		
		if (!m_pLens->getAlphaVectorDerivatives(newTheta, mxx, myy, mxy))
		{
			setErrorString(m_pLens->getErrorString());
			return false;
		}

		mxx = (1.0-fraction*mxx);
		myy = (1.0-fraction*myy);
		mxy = -fraction*mxy;

		double denom = mxx*myy-mxy*mxy;
		double dx = (myy*diffBeta.getX() - mxy*diffBeta.getY())/denom;
		double dy = (-mxy*diffBeta.getX() + mxx*diffBeta.getY())/denom;
		Vector2D<double> newBeta;

		newTheta += grale::Vector2D<double>(dx, dy);
		if (!m_pLens->traceTheta(m_Ds, m_Dds, newTheta, &newBeta))
		{
			setErrorString(std::string("Unable to trace a point to the source plane: ") + m_pLens->getErrorString());
			return false;
		}

		diffBeta = beta - newBeta;

		if (diffBeta.getLength() < prevDiff)
		{
			prevDiff = diffBeta.getLength();
			closestPoint = newTheta;
		}
	}

	//printf("%g\n",prevDiff/ANGLE_ARCSEC);

	theta = closestPoint;
	return true;
}

bool BackProjectedImage::calculateBeta(grale::Vector2D<double> theta, grale::Vector2D<double> &beta)
{
	if (!m_pLens)
	{
		setErrorString("Not initialized");
		return false;
	}	

	Vector2D<double> newBeta;
	
	if (!m_pLens->traceTheta(m_Ds, m_Dds, theta, &newBeta))
	{
		setErrorString(std::string("Unable to trace a point to the source plane: ") + m_pLens->getErrorString());
		return false;
	}

	beta = newBeta;
	return true;
}

GridFunction *BackProjectedImage::createSourceGrid(Vector2D<double> bottomLeft, Vector2D<double> topRight,
	                                           int width, int height)
{
	if (!m_pLens)
	{
		setErrorString("Not initialized");
		return false;
	}	
	
	GridFunction *pGrid = new GridFunction(m_betaTriangles, bottomLeft, topRight, width, height, false, true, 4);
	return pGrid;
}

GridFunction *BackProjectedImage::createImageGrid(Vector2D<double> bottomLeft, Vector2D<double> topRight,
                                                  int width, int height)
{
	if (!m_pLens)
	{
		setErrorString("Not initialized");
		return false;
	}	
	
	GridFunction *pGrid = new GridFunction(m_thetaTriangles, bottomLeft, topRight, width, height, false, true, 4);
	return pGrid;
}

