/*

  This file is a part of GRALEPS, a point selection utility to help
  create input data for gravitational lens inversion using GRALE.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "imagestack.h"
#include "imagesdatawidget.h"
#include "backprojectedimage.h"
#include "sourceandimagewidget.h"
#include "imagestackcontrolwidget.h"
#include <grale/imagesdata.h>
#include <grale/gravitationallens.h>
#include <q3progressbar.h>
#include <qapplication.h>
#include <stdio.h>
#include <memory>

#include <iostream>

using namespace grale;
using namespace std;

ImagesDataWidget::ImagesDataWidget(QWidget *pParent) : Q3Table(pParent)
{
	setReadOnly(true);

	setNumRows(1);
	setNumCols(1);
	setItem(0, 0, new Q3TableItem(this, Q3TableItem::Never, QString("No images data loaded")));
}

ImagesDataWidget::~ImagesDataWidget()
{
}

bool ImagesDataWidget::init(double Ds, double Dds)
{
	if (m_initialized)
	{
		setErrorString("Already initialized");
		return false;
	}

	m_Ds = Ds;
	m_Dds = Dds;

	m_initialized = true;
	return true;
}

void ImagesDataWidget::clear()
{
	if (!m_initialized)
		return;

	clearWidgets();

	for (int i = 0 ; i < m_imagesDataSet.size() ; i++)
		delete m_imagesDataSet[i];
	m_imagesDataSet.resize(0);
	
	for (int i = 0 ; i < m_imageStacks.size() ; i++)
		delete m_imageStacks[i];
	m_imageStacks.resize(0);
		
	setNumRows(1);
	setNumCols(1);
	setItem(0, 0, new Q3TableItem(this, Q3TableItem::Never, QString("No images data loaded")));

	m_initialized = false;
}

void ImagesDataWidget::clearWidgets()
{
	if (!m_initialized)
		return;

	while (!m_imageStackWidgets.empty())
	{
		delete *(m_imageStackWidgets.begin());
		m_imageStackWidgets.pop_front();
	}
}

bool ImagesDataWidget::addImagesData(const std::string &fileName, const std::string &name, const GravitationalLens &lens)
{
	if (!m_initialized)
	{
		setErrorString("Not initialized");
		return false;
	}

	auto_ptr<ImagesData> pImgDat(new ImagesData());

	if (!pImgDat->load(fileName))
	{
		setErrorString(std::string("Unable to load images data: ") + pImgDat->getErrorString());
		return false;
	}

	int numImages = pImgDat->getNumberOfImages();

	if (numImages < 1)
	{
		setErrorString(std::string("Invalid number of images in ") + fileName);
		return false;
	}

	if (m_imagesDataSet.size() > 0)
	{
		if (numImages != m_imagesDataSet[0]->getNumberOfImages())
		{
			setErrorString("File contains a different amount of images than the existing images data sets");
			return false;
		}
	}
	else
	{
		setNumRows(numImages);

		m_imageStacks.resize(numImages);

		for (int i = 0 ; i < numImages ; i++)
			m_imageStacks[i] = new ImageStack();
	}
	
	ImagesData *pDat = pImgDat.release();
	
	m_imagesDataSet.push_back(pDat);

	setNumCols(m_imagesDataSet.size());

	int newCol = m_imagesDataSet.size()-1;
	
	Q3ProgressBar progress(numImages, this, "progress", Qt::WType_Dialog|Qt::WShowModal);

	std::string caption = std::string("Importing images from ") + fileName;

	progress.resize(640,40);
	progress.setCaption(caption.c_str());
	progress.show();
	progress.setProgress(0);

	for (int i = 0 ; i < numImages ; i++)
	{
		qApp->processEvents();

		char str[256];

		snprintf(str, 255, "%d", pDat->getNumberOfImagePoints(i));
		setItem(i, newCol, new Q3TableItem(this, Q3TableItem::Never, str));

		BackProjectedImage *pImage = new BackProjectedImage();

		// TODO: check for errors here
		pImage->init(*pDat, i, lens, m_Ds, m_Dds);

		m_imageStacks[i]->addImage(pImage, true);

		progress.setProgress(i+1);
	}

	horizontalHeader()->setLabel(newCol, name.c_str());

	if (!m_imageStackWidgets.empty()) // render it
	{
		Q3ProgressBar progress(numImages, this, "progress", Qt::WType_Dialog|Qt::WShowModal);

		std::string caption = std::string("Creating source and image views");

		progress.resize(640,40);
		progress.setCaption(caption.c_str());
		progress.show();
		progress.setProgress(0);

		setCommonSourceRegion();
	
		for (int i = 0 ; i < numImages ; i++)
		{
			int xRes, yRes;
			
			qApp->processEvents();

			calculateResolution(m_imageStacks[i]->getSourceBottomLeft(), m_imageStacks[i]->getSourceTopRight(), &xRes, &yRes);
			m_imageStacks[i]->renderSources(xRes, yRes);
			
			calculateResolution(m_imageStacks[i]->getImageBottomLeft(), m_imageStacks[i]->getImageTopRight(), &xRes, &yRes);
			m_imageStacks[i]->renderImages(xRes, yRes);
			
			progress.setProgress(i+1);
		}
	}

	return true;
}

void ImagesDataWidget::createWidgets(QWidget *pParent, ImageStackControlWidget *pGlobalSourceControl,
		   ImageStackControlWidget *pGlobalImageControl, int minRes,
		   const MainWidget *w)

{

	if (!m_initialized)
		return;

	clearWidgets();

	int numImages = m_imageStacks.size();

	Q3ProgressBar progress(numImages, this, "progress", Qt::WType_Dialog|Qt::WShowModal);

	std::string caption = std::string("Creating source and image views");

	progress.resize(640,40);
	progress.setCaption(caption.c_str());
	progress.show();
	progress.setProgress(0);
	
	m_resolution = minRes;
	
	setCommonSourceRegion();
	
	for (int i = 0 ; i < numImages ; i++)
	{
		int xRes, yRes;
		    
		qApp->processEvents();

		m_imageStacks[i]->setImageRegionExtremes();

		calculateResolution(m_imageStacks[i]->getSourceBottomLeft(), m_imageStacks[i]->getSourceTopRight(), &xRes, &yRes);
		//std::cout << "Source: " << xRes << " " << yRes << std::endl;
		m_imageStacks[i]->renderSources(xRes, yRes);
		
		calculateResolution(m_imageStacks[i]->getImageBottomLeft(), m_imageStacks[i]->getImageTopRight(), &xRes, &yRes);
		//std::cout << "Image: " << xRes << " " << yRes << std::endl;
		m_imageStacks[i]->renderImages(xRes, yRes);

		SourceAndImageWidget *pSrcImg = new SourceAndImageWidget(pParent, m_imageStacks[i], i, w);
		
		char str[256];

		snprintf(str, 255, "Image number %d", (i+1));
		str[255] = 0;
		
		pSrcImg->setCaption(str);
		pSrcImg->show();

		emit signalWidgetCreated(pSrcImg);

		QObject::connect(pGlobalSourceControl, SIGNAL(signalGray(int, double, double)), pSrcImg, SLOT(slotSourceGray(int, double, double)));
		QObject::connect(pGlobalImageControl, SIGNAL(signalGray(int, double, double)), pSrcImg, SLOT(slotImageGray(int, double, double)));
		QObject::connect(pGlobalSourceControl, SIGNAL(signalRGB(int, int, int, double, double, double, double, double, double)), pSrcImg, SLOT(slotSourceRGB(int, int, int, double, double, double, double, double, double)));
		QObject::connect(pGlobalImageControl, SIGNAL(signalRGB(int, int, int, double, double, double, double, double, double)), pSrcImg, SLOT(slotImageRGB(int, int, int, double, double, double, double, double, double)));

		m_imageStackWidgets.push_back(pSrcImg);
		
		progress.setProgress(i+1);
	}
}

void ImagesDataWidget::setCommonSourceRegion()
{
	int numImages = m_imageStacks.size();
	double xMin, yMin, xMax, yMax;

	xMin = 1e30;
	yMin = 1e30;
	xMax = -1e30;
	yMax = -1e30;
	
	for (int i = 0 ; i < numImages ; i++)
	{
		m_imageStacks[i]->setSourceRegionExtremes();

		xMin = MIN(xMin, m_imageStacks[i]->getSourceBottomLeft().getX());
		yMin = MIN(yMin, m_imageStacks[i]->getSourceBottomLeft().getY());
		xMax = MAX(xMax, m_imageStacks[i]->getSourceTopRight().getX());
		yMax = MAX(yMax, m_imageStacks[i]->getSourceTopRight().getY());
	}

	m_sourceBottomLeft = Vector2D<double>(xMin, yMin);
	m_sourceTopRight = Vector2D<double>(xMax, yMax);

	for (int i = 0 ; i < numImages ; i++)
		m_imageStacks[i]->setSourceRegion(m_sourceBottomLeft, m_sourceTopRight);
}

void ImagesDataWidget::calculateResolution(grale::Vector2D<double> bottomLeft, grale::Vector2D<double> topRight, int *xRes, int *yRes)
{
	double xDiff = topRight.getX() - bottomLeft.getX();
	double yDiff = topRight.getY() - bottomLeft.getY();

	double frac = xDiff/yDiff;

	if (frac > 1.0)
	{
		*xRes = m_resolution;
		*yRes = (int)(m_resolution/frac+0.5);
	}
	else
	{
		*yRes = m_resolution;
		*xRes = (int)(m_resolution*frac+0.5);
	}
}

