/*

  This file is a part of GRALEPS, a point selection utility to help
  create input data for gravitational lens inversion using GRALE.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "imagestack.h"
#include "backprojectedimage.h"
#include <grale/gridfunction.h>
#include <grale/vector2d.h>

using namespace grale;

ImageStack::ImageStack()
{
	m_renderAllImages = true;
	m_renderAllSources = true;
}

ImageStack::~ImageStack()
{
	clear();
}

void ImageStack::clear()
{
	for (int i = (int)m_imageStack.size() - 1 ; i >= 0 ; i--)
		removeImage(i);
}

void ImageStack::addImage(BackProjectedImage *pImage, bool deleteImage)
{
	m_imageStack.push_back(std::pair<BackProjectedImage *, bool> (pImage, deleteImage));
	m_sourceGrids.push_back(0);
	m_imageGrids.push_back(0);
}

bool ImageStack::removeImage(int index)
{
	if (index < 0 || index >= m_imageStack.size())
	{
		setErrorString("Invalid image index");
		return false;
	}

	if (m_imageStack[index].second && m_imageStack[index].first != 0)
		delete m_imageStack[index].first;
	if (m_sourceGrids[index] != 0)
		delete m_sourceGrids[index];
	if (m_imageGrids[index] != 0)
		delete m_imageGrids[index];

	for (int i = index+1 ; i < m_imageStack.size() ; i++)
	{
		m_imageStack[i-1] = m_imageStack[i];
		m_sourceGrids[i-1] = m_sourceGrids[i];
		m_imageGrids[i-1] = m_imageGrids[i];
	}

	int newSize = m_imageStack.size() - 1;
	
	m_imageStack.resize(newSize);
	m_sourceGrids.resize(newSize);
	m_imageGrids.resize(newSize);
	
	return true;
}

void ImageStack::setSourceRegion(grale::Vector2D<double> bottomLeft, grale::Vector2D<double> topRight)
{
	m_sourceTopRight = topRight;
	m_sourceBottomLeft = bottomLeft;
	m_renderAllSources = true;
}

bool ImageStack::setSourceRegion(int index)
{
	if (index < 0 || index >= m_imageStack.size())
	{
		setErrorString("Invalid image index");
		return false;
	}

	m_sourceTopRight = m_imageStack[index].first->getSourceTopRight();
	m_sourceBottomLeft = m_imageStack[index].first->getSourceBottomLeft();
	m_renderAllSources = true;
	
	return true;
}

void ImageStack::setSourceRegionExtremes()
{
	if (m_imageStack.size() <= 0)
	{
		m_sourceTopRight = Vector2D<double>(0, 0);
		m_sourceBottomLeft = Vector2D<double>(0, 0);
	}
	else
	{
		double maxx = m_imageStack[0].first->getSourceTopRight().getX();
		double maxy = m_imageStack[0].first->getSourceTopRight().getY();
		double minx = m_imageStack[0].first->getSourceBottomLeft().getX();
		double miny = m_imageStack[0].first->getSourceBottomLeft().getY();

		for (int i = 1 ; i < m_imageStack.size() ; i++)
		{
			maxx = MAX(maxx, m_imageStack[0].first->getSourceTopRight().getX());
			maxy = MAX(maxy, m_imageStack[0].first->getSourceTopRight().getY());
			minx = MIN(minx, m_imageStack[0].first->getSourceBottomLeft().getX());
			miny = MIN(miny, m_imageStack[0].first->getSourceBottomLeft().getY());
		}

		m_sourceTopRight = Vector2D<double>(maxx, maxy);
		m_sourceBottomLeft = Vector2D<double>(minx, miny);
	}
	m_renderAllSources = true;
}

void ImageStack::setImageRegion(grale::Vector2D<double> bottomLeft, grale::Vector2D<double> topRight)
{
	m_imageTopRight = topRight;
	m_imageBottomLeft = bottomLeft;
	m_renderAllImages = true;
}

bool ImageStack::setImageRegion(int index)
{
	if (index < 0 || index >= m_imageStack.size())
	{
		setErrorString("Invalid image index");
		return false;
	}

	m_imageTopRight = m_imageStack[index].first->getImageTopRight();
	m_imageBottomLeft = m_imageStack[index].first->getImageBottomLeft();
	m_renderAllImages = true;
}

void ImageStack::setImageRegionExtremes()
{
	if (m_imageStack.size() <= 0)
	{
		m_imageTopRight = Vector2D<double>(0, 0);
		m_imageBottomLeft = Vector2D<double>(0, 0);
	}
	else
	{
		double maxx = m_imageStack[0].first->getImageTopRight().getX();
		double maxy = m_imageStack[0].first->getImageTopRight().getY();
		double minx = m_imageStack[0].first->getImageBottomLeft().getX();
		double miny = m_imageStack[0].first->getImageBottomLeft().getY();

		for (int i = 1 ; i < m_imageStack.size() ; i++)
		{
			maxx = MAX(maxx, m_imageStack[0].first->getImageTopRight().getX());
			maxy = MAX(maxy, m_imageStack[0].first->getImageTopRight().getY());
			minx = MIN(minx, m_imageStack[0].first->getImageBottomLeft().getX());
			miny = MIN(miny, m_imageStack[0].first->getImageBottomLeft().getY());
		}

		m_imageTopRight = Vector2D<double>(maxx, maxy);
		m_imageBottomLeft = Vector2D<double>(minx, miny);
	}
	m_renderAllImages = true;
}

bool ImageStack::renderSources(int width, int height)
{
	return renderInternal(m_sourceGrids, m_sourceBottomLeft, m_sourceTopRight, m_renderAllSources, width, height, true);
}

bool ImageStack::renderImages(int width, int height)
{
	return renderInternal(m_imageGrids, m_imageBottomLeft, m_imageTopRight, m_renderAllImages, width, height, false);
}

bool ImageStack::renderInternal(std::vector<grale::GridFunction *> &grids, grale::Vector2D<double> bottomLeft,
                                grale::Vector2D<double> topRight, bool renderAll, int width, int height, bool source)
{
	if (width < 1 || height < 1 || width > 30000 || height > 30000)
	{
		setErrorString("Invalid width or height");
		return false;
	}
	
	for (int i = 0 ; i < grids.size() ; i++)
	{
		if (grids[i] == 0 || renderAll)
		{
			GridFunction *pGridFunction = 0;

			if (source)
				pGridFunction = m_imageStack[i].first->createSourceGrid(bottomLeft, topRight, width, height);
			else
				pGridFunction = m_imageStack[i].first->createImageGrid(bottomLeft, topRight, width, height);

			if (pGridFunction == 0)
			{
				setErrorString(std::string("Unable to render: ") + m_imageStack[i].first->getErrorString());
				return false;
			}

			if (grids[i])
				delete grids[i];

			grids[i] = pGridFunction;
		}
	}

	if (source)
		m_renderAllSources = false;
	else
		m_renderAllImages = false;

	return true;
}

const GridFunction *ImageStack::getSourceGrid(int index) const
{
	if (index < 0 || index >= m_imageStack.size())
	{
		setErrorString("Invalid image index");
		return 0;
	}

	return m_sourceGrids[index];
}

const GridFunction *ImageStack::getImageGrid(int index) const
{
	if (index < 0 || index >= m_imageStack.size())
	{
		setErrorString("Invalid image index");
		return 0;
	}

	return m_imageGrids[index];
}

bool ImageStack::calculateTheta(int index, grale::Vector2D<double> beta, grale::Vector2D<double> &theta, int numIterations)
{
	if (index < 0 || index >= m_imageStack.size())
	{
		setErrorString("Invalid image index");
		return false;
	}

	if (!m_imageStack[index].first->calculateTheta(beta, theta, numIterations))
	{
		setErrorString(m_imageStack[index].first->getErrorString());
		return false;
	}

	return true;
}

bool ImageStack::calculateBeta(int index, grale::Vector2D<double> theta, grale::Vector2D<double> &beta)
{
	if (index < 0 || index >= m_imageStack.size())
	{
		setErrorString("Invalid image index");
		return false;
	}

	if (!m_imageStack[index].first->calculateBeta(theta, beta))
	{
		setErrorString(m_imageStack[index].first->getErrorString());
		return false;
	}

	return true;
}

