/*

  This file is a part of GRALEPS, a point selection utility to help
  create input data for gravitational lens inversion using GRALE.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "imagestackwidget.h"
#include "imagestack.h"
#include "mainwidget.h"
#include <grale/gridfunction.h>
#include <qpainter.h>
#include <qpixmap.h>
#include <qimage.h>
//Added by qt3to4:
#include <QMouseEvent>
#include <iostream>

using namespace grale;

ImageStackWidget::ImageStackWidget(QWidget *pParent, ImageStack *pImageStack, bool source, int imageIndex, const MainWidget *w) : Q3ScrollView(pParent, 0, Qt::WNoAutoErase|Qt::WPaintClever|Qt::WStaticContents), m_pMainWidget(w)
{
	m_pImageStack = pImageStack;
	m_source = source;
	m_grayScale = false;
	m_grayIndex = 0;
	m_redIndex = 0;
	m_greenIndex = 1;
	m_blueIndex = 2;

	for (int i = 0 ; i < 3 ; i++)
	{
		m_rgbMin[i] = 0;
		m_rgbMax[i] = 1;
	}
	m_grayMin = 0;
	m_grayMax = 1;

	m_imageIndex = imageIndex;
}

ImageStackWidget::~ImageStackWidget()
{
}

void ImageStackWidget::drawContents(QPainter *pPainter, int clipx, int clipy, int clipw, int cliph)
{
	QImage image(clipw, cliph, 32);

	int xstart = clipx;
	int ystart = clipy;
	int xstop = clipx+clipw;
	int ystop = clipy+cliph;

	image.fill(0);

	if (m_grayScale)
	{
		const GridFunction *pGrid;
		if (m_source)
			pGrid = m_pImageStack->getSourceGrid(m_grayIndex);
		else
			pGrid = m_pImageStack->getImageGrid(m_grayIndex);

		if (pGrid)
		{
			int gridWidth = pGrid->getWidth();
			int gridHeight = pGrid->getHeight();

			if (gridWidth != contentsWidth() || gridHeight != contentsHeight())
			{
				resizeContents(gridWidth, gridHeight);
			}
			else
			{
				if (xstop > gridWidth)
					xstop = gridWidth;
				if (ystop > gridHeight)
					ystop = gridHeight;

				int x,y,i,j;

				for (x = xstart, i = 0 ; x < xstop ; x++, i++)
				{
					for (y = ystart, j = 0 ; y < ystop ; y++, j++)
					{
						double z = pGrid->getValue(x, gridHeight-1-y);
						int val = (int)(((z-m_grayMin)/(m_grayMax-m_grayMin))*256.0);

						if (val < 0)
							val = 0;
						else if (val > 255)
							val = 255;

						image.setPixel(i, j, qRgb(val, val, val));
					}
				}
			}
		}
	}
	else
	{
		const GridFunction *pGrids[3];
		if (m_source)
		{
			pGrids[0] = m_pImageStack->getSourceGrid(m_redIndex);
			pGrids[1] = m_pImageStack->getSourceGrid(m_greenIndex);
			pGrids[2] = m_pImageStack->getSourceGrid(m_blueIndex);
		}
		else
		{
			pGrids[0] = m_pImageStack->getImageGrid(m_redIndex);
			pGrids[1] = m_pImageStack->getImageGrid(m_greenIndex);
			pGrids[2] = m_pImageStack->getImageGrid(m_blueIndex);
		}

		int gridWidth = 0; 
		int gridHeight = 0;

		for (int i = 2 ; i >= 0 ; i--)
		{
			if (pGrids[i])
			{
				gridWidth = pGrids[i]->getWidth();
				gridHeight = pGrids[i]->getHeight();
			}
		}

		if (gridWidth > 0 && gridHeight > 0)
		{
			if (gridWidth != contentsWidth() || gridHeight != contentsHeight())
			{
				resizeContents(gridWidth, gridHeight);
			}
			else
			{
				if (xstop > gridWidth)
					xstop = gridWidth;
				if (ystop > gridHeight)
					ystop = gridHeight;

				int x,y,i,j;

				for (x = xstart, i = 0 ; x < xstop ; x++, i++)
				{
					for (y = ystart, j = 0 ; y < ystop ; y++, j++)
					{
						int val[3] = {0, 0, 0};
						
						for (int k = 0 ; k < 3 ; k++)
						{
							if (pGrids[k])
							{
								double z = pGrids[k]->getValue(x, gridHeight-1-y);
								int v = (int)(((z-m_rgbMin[k])/(m_rgbMax[k]-m_rgbMin[k]))*256.0);

								if (v < 0)
									v = 0;
								else if (v > 255)
									v = 255;

								val[k] = v;
							}
						}

						image.setPixel(i, j, qRgb(val[0], val[1], val[2]));
					}
				}
			}
		}
	}
	QPixmap pixmap(image);

	pPainter->drawPixmap(clipx, clipy, pixmap);

	std::list<std::pair<Vector2D<double>, QRgb> > points, points2;
	std::list<std::pair<Vector2D<double>, QRgb> >::const_iterator it;

	m_pMainWidget->getPointGroupInfo(m_imageIndex, points);
	m_pMainWidget->getTimeDelayInfo(m_imageIndex, points2);
	
	// TODO: use another gridfunction?
	const GridFunction *pGrid = 0;
	if (m_source)
		pGrid = m_pImageStack->getSourceGrid(0);
	else
		pGrid = m_pImageStack->getImageGrid(0);

	if (!pGrid)
		return;
	
	for (it = points.begin() ; it != points.end() ; ++it)
	{
		Vector2D<double> point;

		if (m_source)
			m_pImageStack->calculateBeta(0, (*it).first, point);
		else
			point = (*it).first;

		IntVector2D pixel = pGrid->getPosition(point);

		if (pixel.getX() >= 0 && pixel.getY() >= 0 && pixel.getX() < pGrid->getWidth() && pixel.getY() < pGrid->getHeight())
		{
			int xpos = pixel.getX();
			int ypos = pGrid->getHeight()-1-pixel.getY();

			pPainter->setPen(QPen(QColor((*it).second), 2));
			pPainter->drawLine(xpos-4,ypos-4,xpos+4,ypos+4);
			pPainter->drawLine(xpos-4,ypos+4,xpos+4,ypos-4);
		}
	}

	for (it = points2.begin() ; it != points2.end() ; ++it)
	{
		Vector2D<double> point;

		if (m_source)
			m_pImageStack->calculateBeta(0, (*it).first, point);
		else
			point = (*it).first;

		IntVector2D pixel = pGrid->getPosition(point);

		if (pixel.getX() >= 0 && pixel.getY() >= 0 && pixel.getX() < pGrid->getWidth() && pixel.getY() < pGrid->getHeight())
		{
			int xpos = pixel.getX();
			int ypos = pGrid->getHeight()-1-pixel.getY();

			pPainter->setPen(QPen(QColor((*it).second), 2));
			pPainter->drawEllipse(xpos-5, ypos-5, 11, 11);
		}
	}
}

void ImageStackWidget::contentsMouseReleaseEvent(QMouseEvent *pEvent)
{
	const GridFunction *pGrid;
	
	if (m_source)
	{
		pGrid = m_pImageStack->getSourceGrid(0);

		if (pGrid)
		{
			Vector2D<double> beta = pGrid->getPosition(pEvent->x(), pGrid->getHeight()-1-pEvent->y());
			Vector2D<double> theta;

			if (m_pImageStack->calculateTheta(0, beta, theta, 200))
			{
				const GridFunction *pImageGrid = m_pImageStack->getImageGrid(0);

				emit signalPointSelected(m_imageIndex, theta.getX(), theta.getY(), (*pImageGrid)(theta));
			}
		}
	}
	else
	{
		pGrid = m_pImageStack->getImageGrid(0);

		if (pGrid)
		{
			Vector2D<double> theta = pGrid->getPosition(pEvent->x(), pGrid->getHeight()-1-pEvent->y());

			emit signalPointSelected(m_imageIndex, theta.getX(), theta.getY(), (*pGrid)(theta));
		}
	}
}

bool ImageStackWidget::setGrayscale(int index)
{
	m_grayIndex = index;
	m_grayScale = true;
	return true;
}

bool ImageStackWidget::setRGB(int Rindex, int Gindex, int Bindex)
{
	m_redIndex = Rindex;
	m_greenIndex = Gindex;
	m_blueIndex = Bindex;
	m_grayScale = false;
	return true;
}

