/*

  This file is a part of GRALEPS, a point selection utility to help
  create input data for gravitational lens inversion using GRALE.

  Copyright (C) 2008-2012 Jori Liesenborgs

  Contact: jori.liesenborgs@gmail.com
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
*/

#include "mainwidget.h"
#include "imagesdatawidget.h"
#include "imagestackcontrolwidget.h"
#include "pointgroupitem.h"
#include "distancedialogimpl.h"
#include "timedelayitem.h"
#include <grale/gravitationallens.h>
#include <grale/constants.h>
#include <grale/randomnumbergenerator.h>
#include <grale/uniformdistribution.h>
#include <grale/imagesdata.h>
#include <serut/fileserializer.h>
#include <qsplitter.h>
#include <q3listview.h>
#include <qworkspace.h>
#include <qmenubar.h>
#include <qstatusbar.h>
#include <q3popupmenu.h>
#include <q3frame.h>
#include <qlayout.h>
#include <q3filedialog.h>
#include <qfileinfo.h>
#include <qmessagebox.h>
#include <qtabwidget.h>
#include <qinputdialog.h>
#include <q3groupbox.h>
#include <qlabel.h>
#include <qpushbutton.h>
#include <qcolordialog.h>
#include <qimage.h>
#include <qpixmap.h>
#include <qinputdialog.h>
//Added by qt3to4:
#include <Q3HBoxLayout>
#include <QKeyEvent>
#include <Q3VBoxLayout>
#include <QCloseEvent>

#include <stdio.h>
#include <iostream>
#include <algorithm>

using namespace grale;

MainWidget::MainWidget()
{
	setCaption("GRALE Point Selection Utility");

	QSplitter *pSplitter = new QSplitter(Qt::Horizontal, this);

	setCentralWidget(pSplitter);

	QSplitter *pLeftSplitter = new QSplitter(Qt::Vertical, pSplitter);
	
	m_pPointGroupWidget = new Q3ListView(pLeftSplitter);
	m_pPointGroupWidget->addColumn("Groups");
	m_pPointGroupWidget->addColumn("Image");
	m_pPointGroupWidget->setRootIsDecorated(true);
	m_pPointGroupWidget->setSorting(-1);

	m_pTimeDelayWidget = new Q3ListView(pLeftSplitter);
	m_pTimeDelayWidget->addColumn("Image");
	m_pTimeDelayWidget->addColumn("Point");
	m_pTimeDelayWidget->addColumn("Delay");
	m_pTimeDelayWidget->addColumn("Colour");
	m_pTimeDelayWidget->setRootIsDecorated(true);
	m_pTimeDelayWidget->setSorting(-1);

	QSplitter *pSplitter2 = new QSplitter(Qt::Vertical, pSplitter);
	
	Q3GroupBox *pGroupBox = new Q3GroupBox(pSplitter2);
	
	m_pWorkspace = new QTabWidget(pGroupBox, 0, Qt::WNoAutoErase);
	
	Q3GroupBox *pGroupBox2 = new Q3GroupBox(pGroupBox);
	
	m_pGlobalSourceControl = new ImageStackControlWidget(pGroupBox2);
	m_pGlobalImageControl = new ImageStackControlWidget(pGroupBox2);
	
	Q3GroupBox *pGroupBox3 = new Q3GroupBox(pSplitter2);
	m_pImagesDataWidget = new ImagesDataWidget(pGroupBox3);

	QLabel *pLabel = new QLabel("Lens: ", pGroupBox3);
	m_pLensWidget = new QLineEdit(pGroupBox3);
	m_pLensWidget->setReadOnly(true);
	QPushButton *pLensButton = new QPushButton("Load", pGroupBox3);

	Q3VBoxLayout *pV0Layout = new Q3VBoxLayout(pGroupBox);
	pV0Layout->addWidget(m_pWorkspace);
	pV0Layout->addWidget(pGroupBox2);
	Q3HBoxLayout *pControlLayout = new Q3HBoxLayout(pGroupBox2);
	pControlLayout->addWidget(m_pGlobalSourceControl);
	pControlLayout->addWidget(m_pGlobalImageControl);

	Q3VBoxLayout *pV1Layout = new Q3VBoxLayout(pGroupBox3);
	pV1Layout->addWidget(m_pImagesDataWidget);
	Q3HBoxLayout *pLensLayout = new Q3HBoxLayout(pV1Layout);
	pLensLayout->addWidget(pLabel);
	pLensLayout->addWidget(m_pLensWidget);
	pLensLayout->addWidget(pLensButton);
	
	QObject::connect(pLensButton, SIGNAL(clicked()), this, SLOT(onLensLoad()));
	QObject::connect(m_pImagesDataWidget, SIGNAL(signalWidgetCreated(QWidget*)), this, SLOT(onImagesDataViewCreated(QWidget*)));
	QObject::connect(m_pPointGroupWidget, SIGNAL(doubleClicked(Q3ListViewItem *, const QPoint &, int)), this, SLOT(onPointGroupAdd()));

	createMenu();
	statusBar();

	show();

	m_pLens = 0;
	m_Ds = 1000.0;
	m_Dds = 1000.0;
	m_modified = false;
	m_resolution = 400;

	m_keyControl = false;
	m_keyShift = false;
}

MainWidget::~MainWidget()
{
	if (m_pLens)
		delete m_pLens;
}

void MainWidget::onPointGroupNew()
{
	if (m_modified)
	{
		int result = QMessageBox::question(this, "Confirm", "Modifications not saved. Continue?", QMessageBox::Yes, QMessageBox::No|QMessageBox::Default|QMessageBox::Escape);
		
		if (result != QMessageBox::Yes)
			return;
		
		m_modified = false;
	}
	
	while (m_pPointGroupWidget->firstChild())
		delete m_pPointGroupWidget->firstChild();
	while (m_pTimeDelayWidget->firstChild())
		delete m_pTimeDelayWidget->firstChild();
}

void MainWidget::onPointGroupSave()
{
	QString fileName = Q3FileDialog::getSaveFileName(QString::null, "*.groupinfo", this, 0, "Save point group info");
	
	if (fileName == QString::null) // cancelled
		return;
	
	QFileInfo fileInfo(fileName);

	if (fileInfo.exists())
	{
		int result = QMessageBox::question(this, "Confirm", "File already exists. Overwrite?", QMessageBox::Yes, QMessageBox::No|QMessageBox::Default|QMessageBox::Escape);

		if (result != QMessageBox::Yes)
			return;
	}

	std::string fname(fileName.ascii());
	serut::FileSerializer fSer;

	if (!fSer.open(fname, serut::FileSerializer::WriteOnly))
	{
		QMessageBox::critical(this, "Error", fSer.getErrorString().c_str(), QMessageBox::Ok, Qt::NoButton);
		return;
	}

	m_modified = false;

	if (!writeGroupInfo(fSer))
	{
		QMessageBox::critical(this, "Error", getErrorString().c_str(), QMessageBox::Ok, Qt::NoButton);
		return;
	}
}

void MainWidget::onPointGroupLoad()
{
	if (m_modified)
	{
		int result = QMessageBox::question(this, "Confirm", "Modifications not saved. Continue?", QMessageBox::Yes, QMessageBox::No|QMessageBox::Default|QMessageBox::Escape);
		
		if (result != QMessageBox::Yes)
			return;
	}

	Q3FileDialog fileDialog("", "*.groupinfo", this, 0, true);
	fileDialog.setCaption("Load group info");

	if (fileDialog.exec() != QDialog::Accepted)
		return;
	
	QFileInfo fileInfo(fileDialog.selectedFile());

	std::string fname(fileDialog.selectedFile().ascii());
	serut::FileSerializer fSer;

	if (!fSer.open(fname, serut::FileSerializer::ReadOnly))
	{
		QMessageBox::critical(this, "Error", fSer.getErrorString().c_str(), QMessageBox::Ok, Qt::NoButton);
		return;
	}

	if (!readGroupInfo(fSer))
	{
		QMessageBox::critical(this, "Error", getErrorString().c_str(), QMessageBox::Ok, Qt::NoButton);
		return;
	}
	
	m_modified = false;
}

void MainWidget::onQuit()
{
	close();
}

void MainWidget::onImagesDataLoad()
{
	if (m_pLens == 0)
		onLensLoad();

	if (m_pLens == 0)
	{
		QMessageBox::critical(this, "Error", "You must load a lens file first", QMessageBox::Ok, Qt::NoButton);
		return;
	}
	
	Q3FileDialog fileDialog("", "*.imgdata", this, 0, true);
	fileDialog.setCaption("Select an images data file to load");

	if (fileDialog.exec() == QDialog::Accepted)
	{
		QString fileName = fileDialog.selectedFile();
		QFileInfo fileInfo(fileName);
		QString name = fileInfo.fileName();
		
		if (m_pImagesDataWidget->getNumberOfDataSets() == 0) // first data set
		{
			DistanceDialogImpl distDlg(this, m_Ds, m_Dds);

			if (distDlg.exec() != QDialog::Accepted)
				return;
			
			m_Ds = distDlg.getDs();
			m_Dds = distDlg.getDds();
			m_pImagesDataWidget->init(m_Ds, m_Dds);
		}

		if (!m_pImagesDataWidget->addImagesData(fileName.ascii(), name.ascii(), *m_pLens))
		{
			QMessageBox::critical(this, "Couldn't add images data", m_pImagesDataWidget->getErrorString().c_str(), QMessageBox::Ok, Qt::NoButton);
		}
	}
}

void MainWidget::onImagesDataClear()
{
	if (m_pImagesDataWidget->getNumberOfDataSets() == 0)
		return;
	
	int result = QMessageBox::question(this, "Confirm", "Remove all images data sets?", QMessageBox::Yes, QMessageBox::No|QMessageBox::Default|QMessageBox::Escape);

	if (result == QMessageBox::Yes)
		m_pImagesDataWidget->clear();
}

void MainWidget::onImagesDataView()
{
	bool ok;
	int resolution = QInputDialog::getInteger("Resolution", "Specify the resolution for the images", m_resolution, 10, 10000, 1, &ok, this);
		
	if (!ok)
		return;

	m_resolution = resolution;
	m_pImagesDataWidget->createWidgets(m_pWorkspace, m_pGlobalSourceControl, m_pGlobalImageControl, 
			                   m_resolution, this);
	m_pWorkspace->setCurrentPage(0);
}

void MainWidget::onImagesDataClearView()
{
	m_pImagesDataWidget->clearWidgets();
}

void MainWidget::onImagesDataViewCreated(QWidget *pW)
{
	m_pWorkspace->addTab(pW, pW->caption());
	QObject::connect(pW, SIGNAL(signalPointSelected(int, double, double, double)), this, SLOT(onPointSelected(int, double, double, double)));
	QObject::connect(this, SIGNAL(signalGroupInfoChanged()), pW, SLOT(slotGroupInfoChanged()));
}

void MainWidget::onPointSelected(int imageIndex, double x, double y, double h)
{
	if (m_keyControl)
	{
		Q3ListViewItem *pItem = m_pPointGroupWidget->selectedItem();

		if (pItem != 0)
		{
			PointGroupItemBase *pBase = (PointGroupItemBase *)pItem;
			PointGroupItemGroup *pGroup = 0;
	
			if (pBase->getType() == PointGroupItemBase::Point)
				pGroup = (PointGroupItemGroup *)pItem->parent();
			else
				pGroup = (PointGroupItemGroup *)pItem;


			addPoint(pGroup, imageIndex, x, y, h);
	
			m_modified = true;
			emit signalGroupInfoChanged();
		}
	}
	if (m_keyShift)
	{
		bool ok;
		char str[256];

		snprintf(str, 255, "(%lf, %lf)", (double)(x/ANGLE_ARCSEC), (double)(y/ANGLE_ARCSEC));

		QString pointStr = QString("Enter time delay for point ") + QString(str);
		double timeDelay = QInputDialog::getDouble("Time delay", pointStr, 0, -1e200, 1e200, 10, &ok);

		if (ok)
		{
			QRgb rgb = QColorDialog::getRgba(qRgb(255,255,255), &ok, this);

			if (ok)
			{	
				addTimeDelay(imageIndex, x, y, h, timeDelay, rgb);

				m_modified = true;
				emit signalGroupInfoChanged();
			}
		}
	}

	m_keyShift = false;
	m_keyControl = false;
	updateStatus();
}

void MainWidget::onPointGroupAdd()
{
	bool ok;
	
	QString name = QInputDialog::getText("Add group", "Enter a group name: ", QLineEdit::Normal, QString::null, &ok);

	if (!ok)
		return;

	QRgb rgb = QColorDialog::getRgba(qRgb(255,255,255), &ok, this);

	if (!ok)
		return;
	
	Q3ListViewItem *pGroup = addGroup(name, rgb);

	m_pPointGroupWidget->setSelected(pGroup, true);

	m_modified = true;
	emit signalGroupInfoChanged();
}

void MainWidget::onLensLoad()
{
	if (m_pImagesDataWidget->getNumberOfDataSets() != 0)
	{
		QMessageBox::critical(this, "Error", "Can't load a new lens file while images data sets are still loaded", QMessageBox::Ok, Qt::NoButton);
		return;
	}
	
	Q3FileDialog fileDialog("", "*.lensdata", this, 0, true);
	fileDialog.setCaption("Select a lens file to load");

	if (fileDialog.exec() == QDialog::Accepted)
	{
		QString fileName = fileDialog.selectedFile();

		GravitationalLens *pLens;
		std::string errstr;
		
		if (!GravitationalLens::load(fileName.ascii(), &pLens, errstr))
		{
			QMessageBox::critical(this, "Error", errstr.c_str(), QMessageBox::Ok, Qt::NoButton);
			return;
		}

		if (m_pLens)
			delete m_pLens;
		m_pLens = pLens;

		m_pLensWidget->setText(fileName);
	}
}

void MainWidget::createMenu()
{
	QMenuBar *pMenu = menuBar();
	Q3PopupMenu *pFile = new Q3PopupMenu(this);
	
	pFile->insertItem("New", this, SLOT(onPointGroupNew()));
	pFile->insertItem("Add group", this, SLOT(onPointGroupAdd()));
	pFile->insertSeparator();
	pFile->insertItem("Import", this, SLOT(onPointGroupImport()));
	pFile->insertItem("Export", this, SLOT(onPointGroupExport()));
	pFile->insertSeparator();
	pFile->insertItem("Load", this, SLOT(onPointGroupLoad()));
	pFile->insertItem("Save", this, SLOT(onPointGroupSave()));
	pFile->insertSeparator();
	pFile->insertItem("Exit", this, SLOT(onQuit()));
	pMenu->insertItem("File", pFile);

	Q3PopupMenu *pImgDat = new Q3PopupMenu(this);

	pImgDat->insertItem("Load", this, SLOT(onImagesDataLoad()));
	pImgDat->insertItem("Clear", this, SLOT(onImagesDataClear()));
	pImgDat->insertItem("View", this, SLOT(onImagesDataView()));
	pImgDat->insertItem("Clear view", this, SLOT(onImagesDataClearView()));
	pMenu->insertItem("Images", pImgDat);
}
				
void MainWidget::closeEvent(QCloseEvent *pEvt)
{
	if (m_modified)
	{
		int result = QMessageBox::question(this, "Confirm", "Modifications not saved. Quit?", QMessageBox::Yes, QMessageBox::No|QMessageBox::Default|QMessageBox::Escape);
		
		if (result == QMessageBox::Yes)
			pEvt->accept();
	}
	else
		pEvt->accept();
}

void MainWidget::getPointGroupInfo(int imageIndex, std::list<std::pair<grale::Vector2D<double>, QRgb> > &points) const
{
	points.clear();
	
	// TODO: make this configurable
	if (1) // all groups
	{
		Q3ListViewItem *pCurItem = m_pPointGroupWidget->firstChild();

		while (pCurItem != 0)
		{
			PointGroupItemGroup *pGroup = (PointGroupItemGroup *)pCurItem;

			Q3ListViewItem *pCurSubItem = pCurItem->firstChild();

			while (pCurSubItem != 0)
			{
				PointGroupItemPoint *pPoint = (PointGroupItemPoint *)pCurSubItem;

				if (pPoint->getImageNumber() == imageIndex)
					points.push_back(std::pair<Vector2D<double>, QRgb>(Vector2D<double>(pPoint->getX(), pPoint->getY()), pGroup->getRgb()));

				pCurSubItem = pCurSubItem->nextSibling();
			}

			pCurItem = pCurItem->nextSibling();
		}
	}
	else // current group only
	{
		Q3ListViewItem *pItem = m_pPointGroupWidget->selectedItem();

		if (pItem == 0)
			return;

		PointGroupItemBase *pBase = (PointGroupItemBase *)pItem;
		PointGroupItemGroup *pGroup = 0;
		
		if (pBase->getType() == PointGroupItemBase::Point)
			pGroup = (PointGroupItemGroup *)pItem->parent();
		else
			pGroup = (PointGroupItemGroup *)pItem;

		Q3ListViewItem *pCurSubItem = pGroup->firstChild();

		while (pCurSubItem != 0)
		{
			PointGroupItemPoint *pPoint = (PointGroupItemPoint *)pCurSubItem;

			if (pPoint->getImageNumber() == imageIndex)
				points.push_back(std::pair<Vector2D<double>, QRgb>(Vector2D<double>(pPoint->getX(), pPoint->getY()), pGroup->getRgb()));

			pCurSubItem = pCurSubItem->nextSibling();
		}
	}
}

void MainWidget::keyPressEvent(QKeyEvent *pEvt)
{
	if (pEvt->key() == Qt::Key_Control)
	{
		pEvt->accept();

		m_keyControl = true;
		updateStatus();
	}
	else if (pEvt->key() == Qt::Key_Shift)
	{
		pEvt->accept();

		m_keyShift = true;
		updateStatus();
	}
}

void MainWidget::keyReleaseEvent(QKeyEvent *pEvt)
{
	if (pEvt->key() == Qt::Key_Control)
	{
		pEvt->accept();

		m_keyControl = false;
		updateStatus();
	}
	else if (pEvt->key() == Qt::Key_Shift)
	{
		pEvt->accept();

		m_keyShift = false;
		updateStatus();
	}
	else if (pEvt->key() == Qt::Key_Delete)
	{
		pEvt->accept();

		if (m_keyShift)
		{
			Q3ListViewItem *pItem = m_pTimeDelayWidget->selectedItem();

			if (pItem != 0)
			{
				Q3ListViewItem *pNextFocus = pItem->nextSibling();

				delete pItem;
				emit signalGroupInfoChanged();

				m_pTimeDelayWidget->setSelected(pNextFocus, true);
			}
		}
		if (m_keyControl)
		{
			Q3ListViewItem *pItem = m_pPointGroupWidget->selectedItem();

			if (pItem != 0)
			{
				PointGroupItemBase *pBase = (PointGroupItemBase *)pItem;
			
				if (pBase->getType() == PointGroupItemBase::Point)
				{
					PointGroupItemPoint *pPoint = (PointGroupItemPoint *)pBase;
					Q3ListViewItem *pNextFocus = pBase->nextSibling();

					if (pNextFocus == 0)
						pNextFocus = pBase->parent();

					delete pPoint;
					emit signalGroupInfoChanged();

					m_pPointGroupWidget->setSelected(pNextFocus, true);
				}
				else
				{
					PointGroupItemGroup *pGroup = (PointGroupItemGroup *)pBase;

					int result = QMessageBox::question(this, "Confirm", "Delete an entire group?", QMessageBox::Yes, QMessageBox::No|QMessageBox::Default|QMessageBox::Escape);

					if (result == QMessageBox::Yes)
					{
						delete pGroup;
						emit signalGroupInfoChanged();
					}

					m_keyControl = false;
					m_keyShift = false;
					updateStatus();
				}
			}
		}
	}
}

void MainWidget::updateStatus()
{
	char str[1024];

	snprintf(str, 1023, "Group: %s | Time delay: %s", ((m_keyControl)?"yes":"no"), ((m_keyShift)?"yes":"no"));
	statusBar()->message(QString(str));
}

#define GROUPINFO_IDENTIFIER 0x464e5247

bool MainWidget::writeGroupInfo(serut::SerializationInterface &si)
{
	int32_t id = GROUPINFO_IDENTIFIER;

	if (!si.writeInt32(id))
	{
		setErrorString(std::string("Couldn't write identifier: ") + si.getErrorString());
		return false;
	}

	int32_t count = 0;
	Q3ListViewItem *pCurItem = m_pPointGroupWidget->firstChild();

	while (pCurItem != 0)
	{
		count++;
		pCurItem = pCurItem->nextSibling();
	}

	if (!si.writeInt32(count))
	{
		setErrorString(std::string("Couldn't write group count: ") + si.getErrorString());
		return false;
	}

	pCurItem = m_pPointGroupWidget->firstChild();
	while (pCurItem != 0)
	{
		PointGroupItemGroup *pGroup = (PointGroupItemGroup *)pCurItem;
		
		std::string name(pGroup->text(0).ascii());
		int32_t rgb[3];
		
		rgb[0] = qRed(pGroup->getRgb());
		rgb[1] = qGreen(pGroup->getRgb());
		rgb[2] = qBlue(pGroup->getRgb());

		if (!si.writeString(name))
		{
			setErrorString(std::string("Couldn't write group name: ") + si.getErrorString());
			return false;
		}

		if (!si.writeInt32s(rgb, 3))
		{
			setErrorString(std::string("Couldn't write group color: ") + si.getErrorString());
			return false;
		}
		
		Q3ListViewItem *pCurSubItem = pCurItem->firstChild();
		int32_t pointCount = 0;
		
		while (pCurSubItem != 0)
		{
			pointCount++;
			pCurSubItem = pCurSubItem->nextSibling();
		}

		if (!si.writeInt32(pointCount))
		{
			setErrorString(std::string("Couldn't write group point count: ") + si.getErrorString());
			return false;
		}

		pCurSubItem = pCurItem->firstChild();
		while (pCurSubItem != 0)
		{
			PointGroupItemPoint *pPoint = (PointGroupItemPoint *)pCurSubItem;
			double xyh[3];
			int32_t imageIndex;
			
			xyh[0] = pPoint->getX();
			xyh[1] = pPoint->getY();
			xyh[2] = pPoint->getHeight();
			imageIndex = pPoint->getImageNumber();
		
			if (!si.writeDoubles(xyh, 3))
			{
				setErrorString(std::string("Couldn't write group point coordinates: ") + si.getErrorString());
				return false;
			}

			if (!si.writeInt32(imageIndex))
			{
				setErrorString(std::string("Couldn't write group point image index: ") + si.getErrorString());
				return false;
			}
			
			pCurSubItem = pCurSubItem->nextSibling();
		}

		pCurItem = pCurItem->nextSibling();
	}
	
	count = 0;
	pCurItem = m_pTimeDelayWidget->firstChild();

	while (pCurItem != 0)
	{
		count++;
		pCurItem = pCurItem->nextSibling();
	}

	if (!si.writeInt32(count))
	{
		setErrorString(std::string("Couldn't write time delay: ") + si.getErrorString());
		return false;
	}

	pCurItem = m_pTimeDelayWidget->firstChild();
	while (pCurItem != 0)
	{
		TimeDelayItem *pTimeDelayItem = (TimeDelayItem *)pCurItem;

		int32_t rgb[3];
		double xyhd[4];
		int32_t imageIndex;
			
		rgb[0] = qRed(pTimeDelayItem->getColour());
		rgb[1] = qGreen(pTimeDelayItem->getColour());
		rgb[2] = qBlue(pTimeDelayItem->getColour());

		xyhd[0] = pTimeDelayItem->getX();
		xyhd[1] = pTimeDelayItem->getY();
		xyhd[2] = pTimeDelayItem->getHeight();
		xyhd[3] = pTimeDelayItem->getTimeDelay();

		imageIndex = pTimeDelayItem->getImageNumber();

		if (!si.writeInt32s(rgb, 3))
		{
			setErrorString(std::string("Couldn't write time delay point color: ") + si.getErrorString());
			return false;
		}
		
		if (!si.writeDoubles(xyhd, 4))
		{
			setErrorString(std::string("Couldn't write time delay point info: ") + si.getErrorString());
			return false;
		}

		if (!si.writeInt32(imageIndex))
		{
			setErrorString(std::string("Couldn't write time delay point image index: ") + si.getErrorString());
			return false;
		}

		pCurItem = pCurItem->nextSibling();
	}
	return true;
}

bool MainWidget::readGroupInfo(serut::SerializationInterface &si)
{
	while (m_pPointGroupWidget->firstChild())
		delete m_pPointGroupWidget->firstChild();
	while (m_pTimeDelayWidget->firstChild())
		delete m_pPointGroupWidget->firstChild();

	int32_t id = GROUPINFO_IDENTIFIER;
	int32_t id2;

	if (!si.readInt32(&id2))
	{
		setErrorString(std::string("Couldn't read identifier: ") + si.getErrorString());
		return false;
	}

	if (id2 != id)
	{
		setErrorString("Read incorrect identifier");
		return false;
	}
	
	int32_t groupCount;

	if (!si.readInt32(&groupCount))
	{
		setErrorString(std::string("Couldn't read group count: ") + si.getErrorString());
		return false;
	}

	for (int32_t g = 0 ; g < groupCount ; g++)
	{
		std::string groupName;
		int32_t rgb[3];
		int32_t numPoints;

		if (!si.readString(groupName))
		{
			setErrorString(std::string("Couldn't read group name: ") + si.getErrorString());
			return false;
		}

		if (!si.readInt32s(rgb, 3))
		{
			setErrorString(std::string("Couldn't read group color: ") + si.getErrorString());
			return false;
		}

		Q3ListViewItem *pGroup = addGroup(groupName.c_str(), qRgb(rgb[0], rgb[1], rgb[2]));

		if (!si.readInt32(&numPoints))
		{
			setErrorString(std::string("Couldn't read number of group points: ") + si.getErrorString());
			return false;
		}

		for (int32_t p = 0 ; p < numPoints ; p++)
		{
			double xyh[3];
			int32_t imageIndex;

			if (!si.readDoubles(xyh, 3))
			{
				setErrorString(std::string("Couldn't read point coordinates: ") + si.getErrorString());
				return false;
			}

			if (!si.readInt32(&imageIndex))
			{
				setErrorString(std::string("Couldn't read point image index: ") + si.getErrorString());
				return false;
			}

			addPoint(pGroup, imageIndex, xyh[0], xyh[1], xyh[2]);
		}
	}
	
	int32_t pointCount;

	if (!si.readInt32(&pointCount))
	{
		setErrorString(std::string("Couldn't read time delay point count: ") + si.getErrorString());
		return false;
	}

	for (int32_t p = 0 ; p < pointCount ; p++)
	{
		int32_t rgb[3];
		double xyhd[4];
		int32_t imageIndex;
			
		if (!si.readInt32s(rgb, 3))
		{
			setErrorString(std::string("Couldn't read time delay point color: ") + si.getErrorString());
			return false;
		}

		if (!si.readDoubles(xyhd, 4))
		{
			setErrorString(std::string("Couldn't read time delay point info: ") + si.getErrorString());
			return false;
		}

		if (!si.readInt32(&imageIndex))
		{
			setErrorString(std::string("Couldn't read time delay point image index: ") + si.getErrorString());
			return false;
		}

		addTimeDelay(imageIndex, xyhd[0], xyhd[1], xyhd[2], xyhd[3], qRgb(rgb[0], rgb[1], rgb[2]));
	}

	return true;
}

#define YUVTORGB(Y,U,V,R,G,B) { \
				float r,g,b; \
						\
				r = 1.164*((float)(Y) - 16.0) + 1.596*((float)(V) - 128.0); \
				g = 1.164*((float)(Y) - 16.0) - 0.813*((float)(V) - 128.0) - 0.391*((float)(U) - 128.0); \
				b = 1.164*((float)(Y) - 16.0) + 2.018*((float)(U) - 128.0); \
							\
				r = (r > 255.0)?255.0:r; \
				g = (g > 255.0)?255.0:g; \
				b = (b > 255.0)?255.0:b; \
				r = (r < 0.0)?0.0:r; \
				g = (g < 0.0)?0.0:g; \
				b = (b < 0.0)?0.0:b; \
						\
				R = (int)r; \
				G = (int)g; \
				B = (int)b; \
			      }


void MainWidget::onPointGroupImport()
{
	Q3FileDialog fileDialog("", "*.imgdata", this, 0, true);
	fileDialog.setCaption("Import group info");

	if (fileDialog.exec() != QDialog::Accepted)
		return;
	
	std::string fname(fileDialog.selectedFile().ascii());

	int count = m_pPointGroupWidget->childCount();

	ImagesData imgDat;

	if (!imgDat.load(fname))
	{
		QMessageBox::critical(this, "Couldn't import group info", imgDat.getErrorString().c_str(), QMessageBox::Ok, Qt::NoButton);
		return;
	}

	RandomNumberGenerator rndGen;
	UniformDistribution dist(144, 256, &rndGen);
	UniformDistribution dist2(0, 256, &rndGen);

	int result = QMessageBox::question(this, "What to import", "Import group and timedelay info (yes) or all image points (no)?", QMessageBox::Yes|QMessageBox::Default|QMessageBox::Escape, QMessageBox::No);
		
	if (result == QMessageBox::Yes)
	{
		for (int g = 0 ; g < imgDat.getNumberOfGroups() ; g++)
		{
			char str[256];
			int Y,U,V;
			int R,G,B;

			snprintf(str, 255, "Group %d\n", count++);
			Y = (int)dist.pickNumber();
			U = (int)dist2.pickNumber();
			V = (int)dist2.pickNumber();

			YUVTORGB(Y,U,V,R,G,B);

			Q3ListViewItem *pGroupItem = addGroup(str, qRgb(R, G, B));

			int numPoints = imgDat.getNumberOfGroupPoints(g);

			for (int i = 0 ; i < numPoints ; i++)
			{
				int imgIndex, pointIndex;

				imgDat.getGroupPointIndices(g, i, &imgIndex, &pointIndex);

				Vector2D<double> point = imgDat.getImagePointPosition(imgIndex, pointIndex);
				double intensity = 0;

				if (imgDat.hasIntensities())
					intensity = imgDat.getImagePointIntensity(imgIndex, pointIndex);

				addPoint(pGroupItem, imgIndex, point.getX(), point.getY(), intensity);
			}
		}

		for (int p = 0 ; p < imgDat.getNumberOfTimeDelays() ; p++)
		{
			int Y,U,V;
			int R,G,B;

			Y = (int)dist.pickNumber();
			U = (int)dist2.pickNumber();
			V = (int)dist2.pickNumber();

			YUVTORGB(Y,U,V,R,G,B);

			int imageIndex, pointIndex;
			double delay;

			imgDat.getTimeDelay(p, &imageIndex, &pointIndex, &delay);

			Vector2D<double> point = imgDat.getImagePointPosition(imageIndex, pointIndex);
			double intensity = 0;

			if (imgDat.hasIntensities())
				intensity = imgDat.getImagePointIntensity(imageIndex, pointIndex);

			addTimeDelay(imageIndex, point.getX(), point.getY(), intensity, delay, qRgb(R, G, B));
		}
	}
	else
	{
		for (int i = 0 ; i < imgDat.getNumberOfImages() ; i++)
		{
			char str[256];
			int Y,U,V;
			int R,G,B;

			snprintf(str, 255, "Group %d\n", count++);
			Y = (int)dist.pickNumber();
			U = (int)dist2.pickNumber();
			V = (int)dist2.pickNumber();

			YUVTORGB(Y,U,V,R,G,B);

			Q3ListViewItem *pGroupItem = addGroup(str, qRgb(R, G, B));

			int numPoints = imgDat.getNumberOfImagePoints(i);

			for (int p = 0 ; p < numPoints ; p++)
			{
				Vector2D<double> point = imgDat.getImagePointPosition(i, p);
				double intensity = 0;

				if (imgDat.hasIntensities())
					intensity = imgDat.getImagePointIntensity(i, p);

				addPoint(pGroupItem, i, point.getX(), point.getY(), intensity);
			}
		}
	}

	m_modified = true;
	emit signalGroupInfoChanged();
}

Q3ListViewItem *MainWidget::addGroup(const QString &name, QRgb rgb)
{
	Q3ListViewItem *pGroup = new PointGroupItemGroup(rgb, m_pPointGroupWidget, name);

	QImage img(10,10,32);
	img.fill(rgb);

	QPixmap pixmap(img);
	pGroup->setPixmap(1, pixmap);

	return pGroup;
}

void MainWidget::addPoint(Q3ListViewItem *pParent, int imageIndex, double x, double y, double h)
{
	char str[256];

	snprintf(str, 255, "(%lf, %lf)", (double)(x/ANGLE_ARCSEC), (double)(y/ANGLE_ARCSEC));
	QString coordString(str);

	snprintf(str, 255, " %d", (imageIndex+1));
	QString imageString(str);
	
	PointGroupItemPoint *newPoint = new PointGroupItemPoint(imageIndex, x, y, h, pParent, coordString, imageString);
	pParent->setOpen(true);
}

void MainWidget::addTimeDelay(int imageIndex, double x, double y, double h, double delay, QRgb rgb)
{
	char str[256];

	snprintf(str, 255, "(%lf, %lf)", (double)(x/ANGLE_ARCSEC), (double)(y/ANGLE_ARCSEC));
	QString coordString(str);

	snprintf(str, 255, " %d", (imageIndex+1));
	QString imageString(str);
	
	snprintf(str, 255, " %g", delay);
	QString delayString(str);

	QImage img(10,10,32);
	img.fill(rgb);

	QPixmap pixmap(img);

	TimeDelayItem *newTimeDelay = new TimeDelayItem(imageIndex, x, y, h, delay, rgb, m_pTimeDelayWidget, imageString, coordString, delayString);
	newTimeDelay->setPixmap(3, pixmap);
}

class PointGroupItemPointWrapper
{
public:
	PointGroupItemPointWrapper(PointGroupItemPoint *pPoint) { m_pPoint = pPoint; }

	PointGroupItemPoint *m_pPoint;
};

bool operator<(PointGroupItemPointWrapper p1, PointGroupItemPointWrapper p2)
{
	if (p1.m_pPoint->getImageNumber() < p2.m_pPoint->getImageNumber())
		return true;
	return false;
}

void MainWidget::onPointGroupExport()
{
	QString fileName = Q3FileDialog::getSaveFileName(QString::null, "*.imgdata", this, 0, "Export point group info");
	
	if (fileName == QString::null) // cancelled
		return;
	
	QFileInfo fileInfo(fileName);
	ImagesData imgDat;

	if (fileInfo.exists())
	{
		int result = QMessageBox::question(this, "Confirm", "File already exists.", "Overwrite", "Merge", "Cancel", 2, 2);

		if (result == 2)
			return;
	
		if (result == 1)
		{
			if (!imgDat.load(fileName.ascii()))
			{
				QMessageBox::critical(this, "Couldn't load images data", imgDat.getErrorString().c_str(), QMessageBox::Ok, Qt::NoButton);
				return;
			}
		}

		if (imgDat.getNumberOfGroups() > 0)
		{
			int result = QMessageBox::question(this, "Confirm", "Images data set already contains group info.", "Clear", "Add", "Cancel", 2, 2);

			if (result == 2)
				return;

			if (result == 0)
				imgDat.clearGroupInfo();
		}

		if (imgDat.getNumberOfTimeDelays() > 0)
		{
			int result = QMessageBox::question(this, "Confirm", "Images data set already contains time delay info.", "Clear", "Add", "Cancel", 2, 2);

			if (result == 2)
				return;

			if (result == 0)
				imgDat.clearTimeDelayInfo();
		}
	}

	if (imgDat.getNumberOfImages() == 0)
	{
		int highestIndex = -1;

		// search for the highest image index

		Q3ListViewItem *pCurItem = m_pPointGroupWidget->firstChild();

		while (pCurItem != 0)
		{
			PointGroupItemGroup *pGroup = (PointGroupItemGroup *)pCurItem;

			Q3ListViewItem *pCurSubItem = pCurItem->firstChild();

			while (pCurSubItem != 0)
			{
				PointGroupItemPoint *pPoint = (PointGroupItemPoint *)pCurSubItem;

				int imageIndex = pPoint->getImageNumber();

				if (imageIndex > highestIndex)
					highestIndex = imageIndex;

				pCurSubItem = pCurSubItem->nextSibling();
			}

			pCurItem = pCurItem->nextSibling();
		}

		pCurItem = m_pTimeDelayWidget->firstChild();

		while (pCurItem != 0)
		{
			TimeDelayItem *pTimeDelayItem = (TimeDelayItem *)pCurItem;
			int imageIndex = pTimeDelayItem->getImageNumber();

			if (imageIndex > highestIndex)
				highestIndex = imageIndex;

			pCurItem = pCurItem->nextSibling();
		}

		imgDat.create(highestIndex+1, true, false); // create room for the image points
	}

	bool hasIntensities = imgDat.hasIntensities();
	bool warning = false;
	QString warningString("");

	Q3ListViewItem *pCurItem = m_pPointGroupWidget->firstChild();

	while (pCurItem != 0)
	{
		PointGroupItemGroup *pGroup = (PointGroupItemGroup *)pCurItem;
		int groupIndex = imgDat.addGroup();
		
		if (groupIndex < 0)
		{
			warning = true;
			warningString += QString("Couldn't add group ") + pGroup->text(0) + QString(": ") + imgDat.getErrorString().c_str() + QString("\n");
		}	
		else
		{
			Q3ListViewItem *pCurSubItem = pCurItem->firstChild();

			// We'll make sure that the image indices are in ascending order. That will
			// make it easier when trying to use a group for time delay info.

			std::vector<PointGroupItemPointWrapper> imagePoints;

			while (pCurSubItem != 0)
			{
				PointGroupItemPoint *pPoint = (PointGroupItemPoint *)pCurSubItem;

				imagePoints.push_back(PointGroupItemPointWrapper(pPoint));
	
				pCurSubItem = pCurSubItem->nextSibling();
			}

			std::sort(imagePoints.begin(), imagePoints.end());

			for (int i = 0 ; i < imagePoints.size() ; i++)
			{
				PointGroupItemPoint *pPoint = imagePoints[i].m_pPoint;
				int pointIndex = -1;
				int imageIndex = pPoint->getImageNumber();

				pointIndex = addImagesDataPoint(imgDat, imageIndex, Vector2D<double>(pPoint->getX(), pPoint->getY()), pPoint->getHeight());

				if (pointIndex < 0)
				{
					warning = true;
					warningString += QString("Couldn't add point ") + pPoint->text(0) + QString(" (image ") + pPoint->text(1) + QString("): ") + QString(imgDat.getErrorString().c_str()) + QString("\n");
				}
				else
				{
					if (!imgDat.addGroupPoint(groupIndex, imageIndex, pointIndex))
					{
						warning = true;
						warningString += QString("Couldn't add group point ") + pPoint->text(0) + QString(" (image ") + pPoint->text(1) + QString("): ") + QString(imgDat.getErrorString().c_str()) + QString("\n");
					}
				}
			}
		}

		pCurItem = pCurItem->nextSibling();
	}

	pCurItem = m_pTimeDelayWidget->firstChild();
	while (pCurItem != 0)
	{
		TimeDelayItem *pTimeDelayItem = (TimeDelayItem *)pCurItem;
		int imageIndex = pTimeDelayItem->getImageNumber();

		int pointIndex = addImagesDataPoint(imgDat, imageIndex, Vector2D<double>(pTimeDelayItem->getX(), pTimeDelayItem->getY()), pTimeDelayItem->getHeight());

		if (pointIndex < 0)
		{
			warning = true;
			warningString += QString("Couldn't add point ") + pTimeDelayItem->text(0) + QString(" (image ") + pTimeDelayItem->text(1) + QString("): ") + QString(imgDat.getErrorString().c_str()) + QString("\n");
		}
		else
		{
			if (!imgDat.addTimeDelayInfo(imageIndex, pointIndex, pTimeDelayItem->getTimeDelay()))
			{
				warning = true;
				warningString += QString("Couldn't add group point ") + pTimeDelayItem->text(0) + QString(" (image ") + pTimeDelayItem->text(1) + QString("): ") + QString(imgDat.getErrorString().c_str()) + QString("\n");
			}
		}
	
		pCurItem = pCurItem->nextSibling();
	}

	if (!imgDat.save(fileName.ascii()))
	{
		QMessageBox::critical(this, "Couldn't export images data", imgDat.getErrorString().c_str(), QMessageBox::Ok, Qt::NoButton);
		return;
	}

	if (warning)
		QMessageBox::warning(this, "Warning", warningString, QMessageBox::Ok, Qt::NoButton);
}

void MainWidget::getTimeDelayInfo(int imageIndex, std::list<std::pair<grale::Vector2D<double>, QRgb> > &points) const
{
	points.clear();

	Q3ListViewItem *pCurItem = m_pTimeDelayWidget->firstChild();

	while (pCurItem != 0)
	{
		TimeDelayItem *pTimeDelayItem = (TimeDelayItem *)pCurItem;

		if (pTimeDelayItem->getImageNumber() == imageIndex)
			points.push_back(std::pair<grale::Vector2D<double>, QRgb>(grale::Vector2D<double>(pTimeDelayItem->getX(), pTimeDelayItem->getY()), pTimeDelayItem->getColour()));

		pCurItem = pCurItem->nextSibling();
	}
}
	
int MainWidget::addImagesDataPoint(grale::ImagesData &imgDat, int imageIndex, grale::Vector2D<double> point, double height)
{
	// check if the point already exists
	
	int numPoints = imgDat.getNumberOfImagePoints(imageIndex);
	int index = -1;

	for (int i = 0 ; i < numPoints && index == -1 ; i++)
	{
		Vector2D<double> p2 = imgDat.getImagePointPosition(imageIndex, i);

		if ((p2.getX() == point.getX()) && (p2.getY() == point.getY()))
		{
			if (!imgDat.hasIntensities())
				index = i;
			else
			{
				if (imgDat.getImagePointIntensity(imageIndex, i) == height)
					index = i;
			}
		}
	}

	if (index == -1) // point doens't exist yet, create a new one
	{
		if (!imgDat.hasIntensities())
			index = imgDat.addPoint(imageIndex, point);
		else
			index = imgDat.addPoint(imageIndex, point, height);
	}

	return index;
}

